/* Copyright 2002-2023 CS GROUP
 * Licensed to CS GROUP (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.tutorials.conversion;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.orekit.bodies.CelestialBodyFactory;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.forces.drag.DragForce;
import org.orekit.forces.drag.IsotropicDrag;
import org.orekit.forces.gravity.HolmesFeatherstoneAttractionModel;
import org.orekit.forces.gravity.potential.GravityFieldFactory;
import org.orekit.forces.gravity.potential.NormalizedSphericalHarmonicsProvider;
import org.orekit.frames.FramesFactory;
import org.orekit.models.earth.ReferenceEllipsoid;
import org.orekit.models.earth.atmosphere.Atmosphere;
import org.orekit.models.earth.atmosphere.NRLMSISE00;
import org.orekit.models.earth.atmosphere.data.CssiSpaceWeatherData;
import org.orekit.orbits.Orbit;
import org.orekit.orbits.OrbitType;
import org.orekit.orbits.PositionAngleType;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.tle.TLE;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.propagation.analytical.tle.generation.FixedPointTleGenerationAlgorithm;
import org.orekit.propagation.conversion.DormandPrince853IntegratorBuilder;
import org.orekit.propagation.conversion.FiniteDifferencePropagatorConverter;
import org.orekit.propagation.conversion.NumericalPropagatorBuilder;
import org.orekit.propagation.numerical.NumericalPropagator;
import org.orekit.propagation.sampling.OrekitFixedStepHandler;
import org.orekit.utils.IERSConventions;
import org.orekit.utils.PVCoordinatesProvider;
import org.orekit.utils.TimeStampedPVCoordinates;


/** Orekit tutorial for TLE conversion.
 *
 * This tutorial shows how to get a numerical propagator from a TLE and back.
 * A propagator converter is used to get a numerical propagator from a single TLE.
 * The fitting is performed on two orbits with 10 points per orbit.
 * Then a TLE is recovered from the initial state of the numerical propagator with an alternative method.
 * Finally, 3 propagators are run on a time span of 2 orbits to see how the differences on the position evolve.
 *
 * @author Pascal Parraud
 */
public class TLEConversion {

    /** Private constructor. */
    private TLEConversion() {
        // Do nothing
    }

    /** Program entry point.
     * @param args program arguments (unused here)
     */
    public static void main(final String[] args) {

        // configure Orekit
        final File home       = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, unzip it in %s and rename it 'orekit-data' for this tutorial to work%n",
                              "orekit-data-master.zip", "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                              home.getAbsolutePath());
            System.exit(1);
        }
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));

        // reference TLE (ISS TLE retrieved from the web)
        final TLE refTle = new TLE("1 25544U 98067A   22316.69632410  .00011847  00000-0  21541-3 0  9992",
                                   "2 25544  51.6435 320.8964 0006852  65.4219  92.4065 15.49965179368218");

        // Gets the propagator to use with the reference TLE (SGP4 or SDP4)
        final Propagator tlePropagator = TLEPropagator.selectExtrapolator(refTle);

        // Retrieves the reference orbit from the initial state of the TLE propagator
        final Orbit refOrbit = tlePropagator.getInitialState().getOrbit();

        // Creates an integrator builder
        final double minStep = 1e-3;
        final double maxStep = 300;
        final double dP      = 1e-3;
        final DormandPrince853IntegratorBuilder integratorBuilder = new DormandPrince853IntegratorBuilder(minStep, maxStep, dP);

        // Creates a numerical propagator builder to fit the samples generated by the TLE propagator
        final NumericalPropagatorBuilder numPropBuilder = new NumericalPropagatorBuilder(refOrbit, integratorBuilder,
                                                                                         PositionAngleType.TRUE, 1.0);
        setUpPropagatorBuilderForceModels(numPropBuilder);

        // Creates the propagator converter
        final FiniteDifferencePropagatorConverter converter = new FiniteDifferencePropagatorConverter(numPropBuilder, 1.0e-3, 1000);

        // Defines time span and steps number for the conversion
        final double duration = 2. * refOrbit.getKeplerianPeriod();
        final int stepNb      = 21;

        // Let's proceed
        final long start = System.currentTimeMillis();
        System.out.println("START NUMERICAL PROPAGATOR FITTING");
        converter.convert(tlePropagator, duration, stepNb);
        System.out.println("END NUMERICAL PROPAGATOR FITTING");
        System.out.println("Duration : " + (System.currentTimeMillis() - start) / 1000 + " s");
        System.out.println("");

        // Gets the numerical propagator that best fits the reference TLE
        final NumericalPropagator numPropagator = (NumericalPropagator) converter.getAdaptedPropagator();

        // Recovers a TLE from the initial state with an alternative method
        final TLE tleOut = TLE.stateToTLE(numPropagator.getInitialState(), refTle, new FixedPointTleGenerationAlgorithm());
        final Propagator outPropagator = TLEPropagator.selectExtrapolator(tleOut);

        System.out.println("INPUT TLE");
        System.out.println(refTle);
        System.out.println("OUTPUT TLE");
        System.out.println(tleOut);
        System.out.println("");

        // Compares the initial states
        System.out.println("Orbit from TLE ref : " + OrbitType.KEPLERIAN.convertType(refOrbit));
        System.out.println("Orbit from NumProp : " + OrbitType.KEPLERIAN.convertType(numPropagator.getInitialState().getOrbit()));
        System.out.println("Orbit from TLE out : " + OrbitType.KEPLERIAN.convertType(outPropagator.getInitialState().getOrbit()));
        System.out.println("");

        // Creates step handlers
        final DeltaPVHandler tleInHandler  = new DeltaPVHandler(tlePropagator);
        final DeltaPVHandler tleOutHandler = new DeltaPVHandler(outPropagator);

        // Adds the step handlers to the numerical propagator
        numPropagator.getMultiplexer().add(60., tleInHandler);
        numPropagator.getMultiplexer().add(60., tleOutHandler);

        // Extrapolates on 2 orbits from the initial reference date
        numPropagator.propagate(refTle.getDate(), refTle.getDate().shiftedBy(duration));

        // Retrieves the differences of position
        final List<TimeStampedPVCoordinates> dPIn  = tleInHandler.getDeltaPV();
        final List<TimeStampedPVCoordinates> dPOut = tleOutHandler.getDeltaPV();

        // Print the results
        System.out.println("         deltaPos (m)");
        System.out.println("dt(')  num-ref  num-out");
        for (int i = 0; i < dPIn.size(); i++) {
            final double d1 = dPIn.get(i).getPosition().getNorm();
            final double d2 = dPOut.get(i).getPosition().getNorm();
            final double dt = dPIn.get(i).getDate().durationFrom(refTle.getDate()) / 60.;
            System.out.format(Locale.US, " %5.1f, %8.3f, %8.3f%n", dt, d1, d2);
        }

    }

    /** Add the forces to the Numerical Propagator Builder.
     * @param propagatorBuilder the propagator builder
     */
    private static void setUpPropagatorBuilderForceModels(final NumericalPropagatorBuilder propagatorBuilder) {
        // Creates the Earth shape
        final ReferenceEllipsoid earth = ReferenceEllipsoid.getWgs84(FramesFactory.getITRF(IERSConventions.IERS_2010, false));

        // Gets the Earth geopotential
        final NormalizedSphericalHarmonicsProvider provider  = GravityFieldFactory.getNormalizedProvider(6, 0);
        final HolmesFeatherstoneAttractionModel geopotential = new HolmesFeatherstoneAttractionModel(earth.getBodyFrame(),
                                                                                                     provider);

        // Creates the Drag
        final CssiSpaceWeatherData swData = new CssiSpaceWeatherData(CssiSpaceWeatherData.DEFAULT_SUPPORTED_NAMES);
        final Atmosphere atmosphere = new NRLMSISE00(swData, CelestialBodyFactory.getSun(), earth);
        final double dragArea = 2500;
        final double dragCoef = 1.2;
        final DragForce drag = new DragForce(atmosphere, new IsotropicDrag(dragArea, dragCoef));

        // Adds the forces to the numerical propagator builder
        propagatorBuilder.addForceModel(geopotential);
        propagatorBuilder.addForceModel(drag);

        // Sets the mass
        propagatorBuilder.setMass(450000.);

    }

    /** Specialized step handler.
     * <p>This class extends the step handler in order to handle PV differences at each step.<p>
     */
    private static class DeltaPVHandler implements OrekitFixedStepHandler {

        /** PV provider . */
        private final PVCoordinatesProvider provider;

        /** PV differences at each step. */
        private final List<TimeStampedPVCoordinates> deltaPV;

        /** Simple constructor.
         * @param provider the {@link PVCoordinatesProvider} to get coordinates
         */
        DeltaPVHandler(final PVCoordinatesProvider provider) {
            this.provider = provider;
            // prepare an empty list of vectors
            this.deltaPV = new ArrayList<TimeStampedPVCoordinates>();
        }

        /** {@inheritDoc} */
        public void handleStep(final SpacecraftState s) {
            // Get the PV from the provider
            final TimeStampedPVCoordinates provided = provider.getPVCoordinates(s.getDate(), s.getFrame());
            // Get the PV from the state
            final TimeStampedPVCoordinates current = s.getPVCoordinates();
            // Get the deltaPV
            final TimeStampedPVCoordinates dPV = new TimeStampedPVCoordinates(s.getDate(),
                                                                              current.getPosition().subtract(provided.getPosition()),
                                                                              current.getVelocity().subtract(provided.getVelocity()));
            // add the current deltaPV
            deltaPV.add(dPV);

        }

        /** Get the PV differences.
         * @return differences of position
         */
        public List<TimeStampedPVCoordinates> getDeltaPV() {
            return deltaPV;
        }

    }
}
